"""
Collection of containers for handling ESM input and output data
"""

# global
import ivy
import ivy_mech
from typing import Dict
from ivy.core.container import Container


def _pad_to_batch_n_time_dims(data, expected_dims):
    # ToDo: remove this once ESM supports arbitrary batch dimensions
    found_dims = len(data.shape)
    if found_dims == expected_dims:
        return data
    elif found_dims < expected_dims:
        return ivy.reshape(data, [1]*(expected_dims - found_dims) + list(data.shape))
    else:
        raise Exception('found more dims {} than expected {}'.format(found_dims, expected_dims))


# noinspection PyMissingConstructor
class ESMCamMeasurement(Container):

    def __init__(self,
                 img_mean: ivy.Array,
                 cam_rel_mat: ivy.Array = None,
                 img_var: ivy.Array = None,
                 validity_mask: ivy.Array = None,
                 pose_mean: ivy.Array = None,
                 pose_cov: ivy.Array = None,
                 dev_str: str = None):
        """
        Create esm image measurement container

        :param img_mean: Camera-relative co-ordinates and image features
                            *[batch_size, timesteps, height, width, 3 + feat]*
        :type: img_mean: array
        :param cam_rel_mat: The pose of the camera relative to the current agent pose. Default is identity matrix
                            *[batch_size, timesteps, 3, 4]*
        :type cam_rel_mat: array, optional
        :param img_var: Image depth and feature variance values, assumed all zero if None.
                        *[batch_size, timesteps, height, width, 1 + feat]*
        :type: img_var: array, optional
        :param validity_mask: Validity mask, for which pixels should be considered. Assumed all valid if None
                                *[batch_size, timesteps, height, width, 1]*
        :type validity_mask: array, optional
        :param pose_mean: The pose of the camera relative to the current agent pose, in rotation vector pose form.
                            Inferred from cam_rel_mat if None. *[batch_size, timesteps, 6]*
        :type pose_mean: array, optional
        :param pose_cov: The convariance of the camera relative pose, in rotation vector form. Assumed all zero if None.
                            *[batch_size, timesteps, 6, 6]*
        :type pose_cov: array, optional
        :param dev_str: Device string to use, default is to use img_mean.
        :type dev_str: str
        """
        if dev_str is None:
            dev_str = ivy.dev_str(img_mean)
        img_mean = _pad_to_batch_n_time_dims(img_mean, 5)
        cont = dict()
        cont['img_mean'] = img_mean
        if cam_rel_mat is None:
            cam_rel_mat = ivy.identity(4, batch_shape=img_mean.shape[0:2], dev_str=dev_str)[..., 0:3, :]
        else:
            cam_rel_mat = _pad_to_batch_n_time_dims(cam_rel_mat, 4)
        cont['cam_rel_mat'] = cam_rel_mat
        if img_var is None:
            img_var = ivy.zeros_like(img_mean, dev_str=dev_str)
        else:
            img_var = _pad_to_batch_n_time_dims(img_var, 5)
        cont['img_var'] = img_var
        if validity_mask is None:
            validity_mask = ivy.ones_like(img_mean[..., 0:1], dev_str=dev_str)
        else:
            validity_mask = _pad_to_batch_n_time_dims(validity_mask, 5)
        cont['validity_mask'] = validity_mask
        if pose_mean is None:
            pose_mean = ivy_mech.mat_pose_to_rot_vec_pose(cam_rel_mat)
        else:
            pose_mean = _pad_to_batch_n_time_dims(pose_mean, 3)
        cont['pose_mean'] = pose_mean
        if pose_cov is None:
            pose_cov = ivy.tile(ivy.expand_dims(ivy.zeros_like(pose_mean, dev_str=dev_str), -1), (1, 1, 1, 6))
        else:
            pose_cov = _pad_to_batch_n_time_dims(pose_cov, 4)
        cont['pose_cov'] = pose_cov
        Container.__init__(self, cont)


# noinspection PyMissingConstructor
class ESMObservation(Container):

    def __init__(self,
                 img_meas: Dict[str, ESMCamMeasurement],
                 agent_rel_mat: ivy.Array,
                 control_mean: ivy.Array = None,
                 control_cov: ivy.Array = None,
                 dev_str: str = None):
        """
        Create esm observation container

        :param img_meas: dict of ESMImageMeasurement objects, with keys for camera names.
        :type: img_meas: Ivy container
        :param agent_rel_mat: The pose of the agent relative to the previous pose, in matrix form
                                *[batch_size, timesteps, 3, 4]*.
        :type agent_rel_mat: array
        :param control_mean: The pose of the agent relative to the previous pose, in rotation vector pose form.
                                Inferred from agent_rel_mat if None. *[batch_size, timesteps, 6]*
        :type control_mean: array, optional
        :param control_cov: The convariance of the agent relative pose, in rotation vector form.
                             Assumed all zero if None. *[batch_size, timesteps, 6, 6]*.
        :type control_cov: array, optional
        :param dev_str: Device string to use, default is to use img_mean.
        :type dev_str: str
        """
        cont = dict()
        if dev_str is None:
            dev_str = ivy.dev_str(agent_rel_mat)
        cont['img_meas'] = Container(img_meas)
        agent_rel_mat = _pad_to_batch_n_time_dims(agent_rel_mat, 4)
        cont['agent_rel_mat'] = agent_rel_mat
        if control_mean is None:
            control_mean = ivy_mech.mat_pose_to_rot_vec_pose(agent_rel_mat)
        else:
            control_mean = _pad_to_batch_n_time_dims(control_mean, 3)
        cont['control_mean'] = control_mean
        if control_cov is None:
            control_cov = ivy.tile(ivy.expand_dims(ivy.zeros_like(control_mean, dev_str=dev_str), -1), (1, 1, 1, 6))
        else:
            control_cov = _pad_to_batch_n_time_dims(control_cov, 4)
        cont['control_cov'] = control_cov
        Container.__init__(self, cont)


# noinspection PyMissingConstructor
class ESMMemory(Container):

    def __init__(self,
                 mean: ivy.Array,
                 var: ivy.Array = None,
                 dev_str: str = None):
        """
        Create esm memory container

        :param mean: The ESM memory feature values *[batch_size, timesteps, omni_height, omni_width, 2 + feat]*
        :type: mean: array
        :param var: The ESM memory feature variance values. All assumed zero if None.
                        *[batch_size, timesteps, omni_height, omni_width, feat]*
        :type: var: array, optional
        :param dev_str: Device string to use, default is to use img_mean.
        :type dev_str: str
        """
        cont = dict()
        mean = _pad_to_batch_n_time_dims(mean, 5)
        cont['mean'] = mean
        if var is None:
            var = ivy.zeros_like(mean, dev_str=dev_str)
        else:
            var = _pad_to_batch_n_time_dims(var, 5)
        cont['var'] = var
        Container.__init__(self, cont)



"""
Implementation of Neural Turing Machine
"""

# global
import ivy
import math
import collections
from ivy.core.container import Container

# local
from ivy.neural_net_stateful.layers import Linear, LSTM


NTMControllerState = collections.namedtuple('NTMControllerState',
                                            ('controller_state', 'read_vector_list', 'w_list', 'usage_indicator', 'M'))


def _expand(x, dim, N):
    return ivy.concatenate([ivy.expand_dims(x, dim) for _ in range(N)], dim)


class NTMCell(ivy.Module):
    def __init__(self, controller, controller_proj, output_proj, output_dim, ctrl_input_size, ctrl_output_size,
                 total_parameter_num, memory_size, memory_vector_dim, read_head_num, write_head_num, v=None, usage=None,
                 addressing_mode='content_and_location', shift_range=1, clip_value=20, init_value=1e-6,
                 sequential_writing=False, retroactive_updates=False, retroactive_discount=0.96, with_erase=True,
                 seed=0):

        # vanilla ntm
        self._memory_size = memory_size
        self._memory_vector_dim = memory_vector_dim
        self._read_head_num = read_head_num
        self._write_head_num = write_head_num
        self._init_value = init_value
        self._addressing_mode = addressing_mode
        self._clip_value = clip_value
        self._output_dim = output_dim
        self._shift_range = shift_range
        self._num_parameters_per_head = self._memory_vector_dim + 1 + 1 + (self._shift_range * 2 + 1) + 1
        self._num_heads = self._read_head_num + self._write_head_num
        ivy.seed(seed)

        # fns + classes
        self._controller = controller
        self._controller_proj = controller_proj
        self._output_proj = output_proj

        # usage
        if usage is not None:
            self._usage = usage
        else:
            self._usage = ivy.zeros([memory_size, ])

        # step
        self._step = 0

        # MERLIN changes
        self._sequential_writing = sequential_writing
        self._retroactive_updates = retroactive_updates
        self._retroactive_discount = retroactive_discount
        self._with_erase = with_erase

        # variables
        ivy.Module.__init__(self, 'cpu')

    def _create_variables(self, dev_str):
        vars_dict = dict()
        wlim = (6 / (2 * self._memory_vector_dim)) ** 0.5
        vars_dict['read_weights'] =\
            dict(zip(['w_' + str(i) for i in range(self._read_head_num)],
                     [ivy.variable(ivy.random_uniform(-wlim, wlim, [self._memory_vector_dim, ], dev_str=dev_str))
                      for _ in range(self._read_head_num)]))
        wlim = (6 / (2 * self._memory_size)) ** 0.5
        vars_dict['write_weights'] =\
            dict(zip(['w_' + str(i) for i in range(self._read_head_num + self._write_head_num)],
                     [ivy.variable(ivy.random_uniform(-wlim, wlim, [self._memory_size, ], dev_str=dev_str))
                      for _ in range(self._read_head_num + self._write_head_num)]))
        vars_dict['memory'] = ivy.variable(
            ivy.ones([self._memory_size, self._memory_vector_dim], dev_str=dev_str) * self._init_value)
        return vars_dict

    def _addressing(self, k, beta, g, s, gamma, prev_M, prev_w):

        # Sec 3.3.1 Focusing by Content

        # Cosine Similarity

        k = ivy.expand_dims(k, axis=2)
        inner_product = ivy.matmul(prev_M, k)
        k_norm = ivy.reduce_sum(k ** 2, axis=1, keepdims=True) ** 0.5
        M_norm = ivy.reduce_sum(prev_M ** 2, axis=2, keepdims=True) ** 0.5
        norm_product = M_norm * k_norm
        K = ivy.squeeze(inner_product / (norm_product + 1e-8))  # eq (6)

        # Calculating w^c

        K_amplified = ivy.exp(ivy.expand_dims(beta, axis=1) * K)
        w_c = K_amplified / ivy.reduce_sum(K_amplified, axis=1, keepdims=True)  # eq (5)

        if self._addressing_mode == 'content':  # Only focus on content
            return w_c

        # Sec 3.3.2 Focusing by Location

        g = ivy.expand_dims(g, axis=1)
        w_g = g * w_c + (1 - g) * prev_w  # eq (7)

        s = ivy.concatenate([s[:, :self._shift_range + 1],
                                 ivy.zeros([s.shape[0], self._memory_size - (self._shift_range * 2 + 1)]),
                                 s[:, -self._shift_range:]], axis=1)
        t = ivy.concatenate([ivy.flip(s, axis=[1]),
                                 ivy.flip(s, axis=[1])], axis=1)
        s_matrix = ivy.stack(
            [t[:, self._memory_size - i - 1:self._memory_size * 2 - i - 1] for i in range(self._memory_size)],
            axis=1)
        w_ = ivy.reduce_sum(ivy.expand_dims(w_g, axis=1) * s_matrix, axis=2)  # eq (8)
        w_sharpen = w_ ** ivy.expand_dims(gamma, axis=1)
        w = w_sharpen / ivy.reduce_sum(w_sharpen, axis=1, keepdims=True)  # eq (9)

        return w

    # Public Methods #
    # ---------------#

    def get_start_state(self, _=None, batch_size=None, dtype_str=None, v=None):
        if v is None:
            v = self.v
        else:
            v = Container(v)
        read_vector_list = [_expand(ivy.tanh(var), dim=0, N=batch_size)
                            for _, var in v.read_weights.to_iterator()]
        w_list = [_expand(ivy.softmax(var), dim=0, N=batch_size)
                  for _, var in v.write_weights.to_iterator()]
        usage_indicator = _expand(self._usage, dim=0, N=batch_size)
        M = _expand(v.memory, dim=0, N=batch_size)
        return NTMControllerState(
            controller_state=self._controller.get_initial_state(batch_shape=(batch_size,)),
            read_vector_list=read_vector_list,
            w_list=w_list,
            usage_indicator=usage_indicator,
            M=M)

    def _forward(self, x, prev_state):
        prev_read_vector_list = prev_state[1]

        controller_input = ivy.concatenate([x] + prev_read_vector_list, axis=1)
        controller_output, controller_state = self._controller(ivy.expand_dims(controller_input, -2),
                                                               initial_state=prev_state[0])
        controller_output = controller_output[..., -1, :]

        parameters = self._controller_proj(controller_output)
        parameters = ivy.clip(parameters, -self._clip_value, self._clip_value)
        head_parameter_list = \
            ivy.split(parameters[:, :self._num_parameters_per_head * self._num_heads], self._num_heads,
                          axis=1)
        erase_add_list = ivy.split(parameters[:, self._num_parameters_per_head * self._num_heads:],
                                       2 * self._write_head_num, axis=1)

        prev_w_list = prev_state[2]
        prev_M = prev_state[4]
        w_list = []
        for i, head_parameter in enumerate(head_parameter_list):
            k = ivy.tanh(head_parameter[:, 0:self._memory_vector_dim])
            beta = ivy.softplus(head_parameter[:, self._memory_vector_dim])
            g = ivy.sigmoid(head_parameter[:, self._memory_vector_dim + 1])
            s = ivy.softmax(
                head_parameter[:, self._memory_vector_dim + 2:self._memory_vector_dim +
                                                              2 + (self._shift_range * 2 + 1)])
            gamma = ivy.softplus(head_parameter[:, -1]) + 1
            w = self._addressing(k, beta, g, s, gamma, prev_M, prev_w_list[i])
            w_list.append(w)

        # Reading (Sec 3.1)

        read_w_list = w_list[:self._read_head_num]
        if self._step == 0:
            usage_indicator = ivy.zeros_like(w_list[0])
        else:
            usage_indicator = prev_state[3] + ivy.reduce_sum(ivy.concatenate(read_w_list, 0))
        read_vector_list = []
        for i in range(self._read_head_num):
            read_vector = ivy.reduce_sum(ivy.expand_dims(read_w_list[i], axis=2) * prev_M, axis=1)
            read_vector_list.append(read_vector)

        # Writing (Sec 3.2)

        prev_wrtie_w_list = prev_w_list[self._read_head_num:]
        w_wr_size = math.ceil(self._memory_size / 2) if self._retroactive_updates else self._memory_size
        if self._sequential_writing:
            batch_size = ivy.shape(x)[0]
            if self._step < w_wr_size:
                w_wr_list = [ivy.tile(ivy.cast(ivy.one_hot(
                    ivy.array([self._step]), w_wr_size), 'float32'),
                    (batch_size, 1))] * self._write_head_num
            else:
                batch_idxs = ivy.expand_dims(ivy.arange(batch_size, 0), -1)
                mem_idxs = ivy.expand_dims(ivy.argmax(usage_indicator[..., :w_wr_size], -1), -1)
                total_idxs = ivy.concatenate((batch_idxs, mem_idxs), -1)
                w_wr_list = [ivy.scatter_nd(total_idxs, ivy.ones((batch_size,)),
                                                (batch_size, w_wr_size))] * self._write_head_num
        else:
            w_wr_list = w_list[self._read_head_num:]
        if self._retroactive_updates:
            w_ret_list = [self._retroactive_discount * prev_wrtie_w[..., w_wr_size:] +
                          (1 - self._retroactive_discount) * prev_wrtie_w[..., :w_wr_size]
                          for prev_wrtie_w in prev_wrtie_w_list]
            w_wrtie_list = [ivy.concatenate((w_wr, w_ret), -1) for w_wr, w_ret in zip(w_wr_list, w_ret_list)]
        else:
            w_wrtie_list = w_wr_list
        M = prev_M
        for i in range(self._write_head_num):
            w = ivy.expand_dims(w_wrtie_list[i], axis=2)
            if self._with_erase:
                erase_vector = ivy.expand_dims(ivy.sigmoid(erase_add_list[i * 2]), axis=1)
                M = M * ivy.ones(ivy.shape(M)) - ivy.matmul(w, erase_vector)
            add_vector = ivy.expand_dims(ivy.tanh(erase_add_list[i * 2 + 1]), axis=1)
            M = M + ivy.matmul(w, add_vector)

        NTM_output = self._output_proj(ivy.concatenate([controller_output] + read_vector_list, axis=1))
        NTM_output = ivy.clip(NTM_output, -self._clip_value, self._clip_value)

        self._step += 1
        return NTM_output, NTMControllerState(
            controller_state=controller_state, read_vector_list=read_vector_list, w_list=w_list,
            usage_indicator=usage_indicator, M=M)

    # Properties #
    # -----------#

    @property
    def _state_size(self):
        return NTMControllerState(
            controller_state=self._controller.state_size[0],
            read_vector_list=[self._memory_vector_dim for _ in range(self._read_head_num)],
            w_list=[self._memory_size for _ in range(self._read_head_num + self._write_head_num)],
            usage_indicator=[self._memory_size],
            M=[self._memory_size, self._memory_vector_dim])

    @property
    def _output_size(self):
        return self._output_dim


class NTM(ivy.Module):

    def __init__(self, input_dim, output_dim, ctrl_output_size, ctrl_layers, memory_size, memory_vector_dim,
                 read_head_num, write_head_num, v=None, usage=None, addressing_mode='content_and_location',
                 shift_range=1, clip_value=20, init_value=1e-6, sequential_writing=False,
                 retroactive_updates=False, retroactive_discount=0.96, with_erase=True):
        ctrl_input_size = read_head_num * memory_vector_dim + input_dim
        num_heads = read_head_num + write_head_num
        num_parameters_per_head = memory_vector_dim + 1 + 1 + (shift_range * 2 + 1) + 1
        total_parameter_num = num_parameters_per_head * num_heads + memory_vector_dim * 2 * write_head_num
        ctrl_v = v.ntm_cell.controller if \
            v is not None and 'ntm_cell' in v and 'controller' in v.ntm_cell else None
        ctrl = LSTM(ctrl_input_size, ctrl_output_size, num_layers=ctrl_layers, v=ctrl_v)
        ctrl_proj_v = v.ntm_cell.controller_proj \
            if v is not None and 'ntm_cell' in v and 'controller_proj' in v.ntm_cell else None
        ctrl_proj = Linear(ctrl_output_size, total_parameter_num, v=ctrl_proj_v)
        out_proj_v = v.ntm_cell.output_proj \
            if v is not None and 'ntm_cell' in v and 'output_proj' in v.ntm_cell else None
        out_proj = Linear(ctrl_output_size + read_head_num * memory_vector_dim, output_dim, v=out_proj_v)

        ntm_v = v.ntm if v is not None and 'ntm' in v else None
        ntm_cell = NTMCell(ctrl, ctrl_proj, out_proj, output_dim, ctrl_input_size, ctrl_output_size,
                           total_parameter_num, memory_size, memory_vector_dim, read_head_num, write_head_num, ntm_v,
                           usage, addressing_mode, shift_range, clip_value, init_value, sequential_writing,
                           retroactive_updates, retroactive_discount, with_erase)
        self._ntm_cell = ntm_cell
        ivy.Module.__init__(self, 'cpu', v=v)

    def _forward(self, inputs, hidden=None):
        inputs_shape = list(inputs.shape)
        batch_shape = inputs_shape[:-2]
        time_dim = inputs_shape[-2]
        inputs = ivy.reshape(inputs, [-1] + list(inputs.shape[-2:]))
        if hidden is None:
            hidden = self._ntm_cell.get_start_state(inputs, inputs.shape[0], inputs.dtype, v=self.v.ntm_cell)
        outputs = []
        for x in ivy.unstack(inputs, 1):
            output, hidden = self._ntm_cell(x, hidden)
            outputs.append(output)
        ret = ivy.stack(outputs, 1)
        return ivy.reshape(ret, batch_shape + [time_dim, -1])



# global
import ivy
import ivy_mech
import ivy_vision

# local
from ivy_memory.geometric.containers import ESMObservation, ESMMemory

MIN_DENOMINATOR = 1e-12


# noinspection PyUnboundLocalVariable
class ESM(ivy.Module):

    def __init__(self, depth_prior=None, feat_prior=None, num_feature_channels=3, smooth_kernel_size=25,
                 threshold_var_factor=0.99, depth_limits=(1e-3, 10.), depth_var_limits=(1e-3, 1e4),
                 feat_var_limits=(1e-3, 1e4), omni_image_dims=(180, 360), smooth_mean=True, depth_buffer=True,
                 stateful=False, device='cpu'):
        # ToDo: make variance fully optional. If not specified,
        #  then do not compute and scatter during function call for better efficiency.

        # features
        self._feat_dim = num_feature_channels

        # hole priors
        if depth_prior is None:
            depth_prior = ivy.ones((1,), dev_str=device) * 1e-2
        if feat_prior is None:
            feat_prior = ivy.ones((num_feature_channels,), dev_str=device) * 0.5
        self._sphere_depth_prior_val = depth_prior
        self._sphere_feat_prior_val = feat_prior

        self._sphere_depth_prior = None
        self._sphere_feat_prior = None

        # value clipping
        self._min_depth = ivy.array(depth_limits[0], dev_str=device)
        self._max_depth = ivy.array(depth_limits[1], dev_str=device)

        # variance clipping
        ang_pix_var_limits = (1e-3, float(omni_image_dims[0]))
        self._min_ang_pix_var = ivy.array(ang_pix_var_limits[0], dev_str=device)
        self._min_depth_var = ivy.array(depth_var_limits[0], dev_str=device)
        self._min_feat_var = ivy.array(feat_var_limits[0], dev_str=device)

        self._ang_pix_prior_var_val = ivy.array(ang_pix_var_limits[1], dev_str=device)
        self._depth_prior_var_val = ivy.array(depth_var_limits[1], dev_str=device)
        self._feat_prior_var_val = ivy.array(feat_var_limits[1], dev_str=device)

        self._threshold_var_factor = ivy.array(threshold_var_factor, dev_str=device)
        self._ang_pix_var_threshold = ivy.array(ang_pix_var_limits[1] * self._threshold_var_factor, dev_str=device)
        self._depth_var_threshold = ivy.array(depth_var_limits[1] * self._threshold_var_factor, dev_str=device)
        self._feat_var_threshold = ivy.array(feat_var_limits[1] * self._threshold_var_factor, dev_str=device)

        # normalization
        self._depth_range = self._max_depth - self._min_depth

        self._min_log_depth_var = ivy.log(self._min_depth_var)
        self._min_log_feat_var = ivy.log(self._min_feat_var)

        max_log_depth_var = ivy.log(self._depth_prior_var_val)
        max_log_feat_var = ivy.log(self._feat_prior_var_val)

        self._log_depth_var_range = max_log_depth_var - self._min_log_depth_var
        self._log_feat_var_range = max_log_feat_var - self._min_log_feat_var

        # sphere image
        self._sphere_img_dims = list(omni_image_dims)
        self._pixels_per_degree = self._sphere_img_dims[0] / 180

        # image smoothing
        self._smooth_kernel_size = smooth_kernel_size
        self._smooth_mean = smooth_mean

        # rendering config
        self._with_depth_buffer = depth_buffer

        # memory
        self._stateful = stateful
        self._memory = None

        # variables
        ivy.Module.__init__(self, device)

    # Helpers #
    # --------#

    def _normalize_mean(self, mean):
        normed_depth = (mean[..., 0:1] - self._min_depth) / (self._depth_range + MIN_DENOMINATOR)
        normed_feat = mean[..., 3:]
        return ivy.concatenate((normed_depth, normed_feat), -1)

    def _normalize_var(self, var):
        normed_depth_var = (ivy.log(var[..., 0:1]) - self._min_log_depth_var) / \
                           (self._log_depth_var_range + MIN_DENOMINATOR)
        normed_feat_var = (ivy.log(var[..., 4:]) - self._min_log_feat_var) / \
                          (self._log_feat_var_range + MIN_DENOMINATOR)
        return ivy.concatenate((normed_depth_var, normed_feat_var), -1)

    @staticmethod
    def _fuse_measurements_with_uncertainty(measurements, measurement_uncertainties, axis):
        measurements_shape = measurements.shape
        batch_shape = measurements_shape[0:axis]
        num_batch_dims = len(batch_shape)
        num_measurements = measurements_shape[axis]

        # BS x 1 x RS
        sum_of_variances = ivy.reduce_sum(measurement_uncertainties, axis, keepdims=True)
        prod_of_variances = ivy.reduce_prod(measurement_uncertainties, axis, keepdims=True)

        # BS x 1 x RS
        new_variance = prod_of_variances / sum_of_variances

        # dim size list of BS x (dim-1) x RS
        batch_slices = [slice(None, None, None)] * num_batch_dims
        concat_lists = \
            [[measurement_uncertainties[tuple(batch_slices + [slice(0, i, None)])]]
             if i != 0 else [] + [measurement_uncertainties[tuple(batch_slices + [slice(i + 1, None, None)])]]
             for i in range(num_measurements)]
        partial_variances_list = [ivy.concatenate(concat_list, axis) for concat_list in concat_lists]

        # dim size list of BS x 1 x RS
        partial_prod_of_variances_list = \
            [ivy.reduce_prod(partial_variance, axis, True) for partial_variance in
             partial_variances_list]

        # BS x dim x RS
        partial_prod_of_variances = ivy.concatenate(partial_prod_of_variances_list, axis)

        # BS x 1 x RS
        new_mean = ivy.reduce_sum((partial_prod_of_variances * measurements) / sum_of_variances, axis, keepdims=True)

        # BS x 1 x RS, BS x 1 x RS
        return new_mean, new_variance

    # Projection Functions #
    # ---------------------#

    def _frame_to_omni_frame_projection(self, cam_rel_poses, cam_rel_mats, uniform_sphere_pixel_coords, cam_coords_f1,
                                        cam_feat_f1, rel_pose_covs, image_var_f1, holes_prior, holes_prior_var,
                                        batch_size, num_timesteps, num_cams, image_dims):
        """
        Project mean and variance values from frame 1 to omni frame 2, using scatter and quantize operations.
        :param cam_rel_poses: Relative pose of camera to agent *[batch_size, n, c, 6]*
        :param cam_rel_mats: Relative transformation matrix from camera to agent *[batch_size, n, c, 3, 4]*
        :param uniform_sphere_pixel_coords: Pixel coords *[batch_size, n, h, w, 3]*
        :param cam_coords_f1: Camera co-ordinates to project *[batch_size, n, c, h_in, w_in, 3]*
        :param cam_feat_f1: Mean feature values to project *[batch_size, n, c, h_in, w_in, f]*
        :param rel_pose_covs: Pose covariances *[batch_size, n, c, 6, 6]*
        :param image_var_f1: Angular pixel, radius and feature variance values to project *[batch_size, n, c, h_in, w_in, 3+f]*
        :param holes_prior: Prior values for holes which arise during the projection *[batch_size, n, h, w, 3+f]*
        :param holes_prior_var: Prior variance for holes which arise during the projection *[batch_size, n, h, w, 3+f]*
        :param batch_size: Size of batch
        :param num_timesteps: Number of timesteps
        :param num_cams: Number of cameras
        :param image_dims: Image dimensions
        :return: Projected mean *[batch_size, 1, h, w, 3+f]* and variance *[batch_size, 1, h, w, 3+f]*
        """

        # cam 1 to cam 2 coords

        cam_coords_f2 = ivy_vision.cam_to_cam_coords(
            ivy_mech.make_coordinates_homogeneous(
                cam_coords_f1, [batch_size, num_timesteps, num_cams] + image_dims), cam_rel_mats,
            [batch_size, num_timesteps, num_cams], image_dims)

        # cam 2 to sphere 2 coords

        sphere_coords_f2 = ivy_vision.cam_to_sphere_coords(cam_coords_f2)
        image_var_f2 = image_var_f1

        # angular pixel coords

        # B x N x C x H x W x 3
        angular_pixel_coords_f2 = \
            ivy_vision.sphere_to_angular_pixel_coords(sphere_coords_f2, self._pixels_per_degree)

        # constant feature projection

        # B x N x C x H x W x (3+F)
        projected_coords_f2 = ivy.concatenate([angular_pixel_coords_f2] + [cam_feat_f1], -1)

        # reshaping to fit quantization dimension requirements

        # B x N x (CxHxW) x (3+F)
        projected_coords_f2_flat = \
            ivy.reshape(projected_coords_f2,
                        [batch_size, num_timesteps, num_cams * image_dims[0] * image_dims[1], -1])

        # B x N x (CxHxW) x (3+F)
        image_var_f2_flat = ivy.reshape(image_var_f2,
                                        [batch_size, num_timesteps, num_cams * image_dims[0] * image_dims[1], -1])

        # quantized result from all scene cameras

        # B x N x OH x OW x (3+F)   # B x N x OH x OW x (3+F)
        return ivy_vision.quantize_to_image(
            pixel_coords=projected_coords_f2_flat[..., 0:2],
            final_image_dims=self._sphere_img_dims,
            feat=projected_coords_f2_flat[..., 2:],
            feat_prior=holes_prior,
            with_db=self._with_depth_buffer,
            pixel_coords_var=image_var_f2_flat[..., 0:2],
            feat_var=image_var_f2_flat[..., 2:],
            pixel_coords_prior_var=holes_prior_var[..., 0:2],
            feat_prior_var=holes_prior_var[..., 2:],
            var_threshold=self._var_threshold,
            uniform_pixel_coords=uniform_sphere_pixel_coords,
            batch_shape=(batch_size, num_timesteps),
            dev_str=self._dev_str)[0:2]

    def _omni_frame_to_omni_frame_projection(self, agent_rel_pose, agent_rel_mat, uniform_sphere_pixel_coords,
                                             sphere_pix_coords_f1, sphere_depth_f1, sphere_feat_f1, agent_rel_pose_cov,
                                             image_var_f1, holes_prior, holes_prior_var, batch_size):
        """
        Project mean and variance values from omni frame 1 to omni frame 2, using scatter and quantize operation.

        :param agent_rel_pose: Relative pose of agent to the previous step *[batch_size, 6]*
        :param agent_rel_mat: Relative transformation matrix of agent to the previous step
                                    *[batch_size, 3, 4]*
        :param uniform_sphere_pixel_coords: Pixel coords *[batch_size, h, w, 3]*
        :param sphere_pix_coords_f1: Pixel co-ordinates to project *[batch_size, h, w, 2]*
        :param sphere_depth_f1: Mean radius values to project *[batch_size, h, w, 1]*
        :param sphere_feat_f1: Mean feature values to project *[batch_size, h, w, f]*
        :param agent_rel_pose_cov: Agent relative pose covariance *[batch_size, 6, 6]*
        :param image_var_f1: Angular pixels, radius and feature variance values to project *[batch_size, h, w, 3+f]*
        :param holes_prior: Prior values for holes which arise during the projection *[batch_size, h, w, 1+f]*
        :param holes_prior_var: Prior variance for holes which arise during the projection
                                *[batch_size, h, w, 3+f]*
        :param batch_size: Size of batch
        :return: Projected mean *[batch_size, h, w, 3+f]* and variance *[batch_size, h, w, 3+f]*
        """

        # Frame 1 #
        # --------#

        # combined

        # B x OH x OW x 3
        angular_pixel_coords_f1 = ivy.concatenate((sphere_pix_coords_f1, sphere_depth_f1), -1)

        # sphere coords

        # B x OH x OW x 3
        sphere_coords_f1 = \
            ivy_vision.angular_pixel_to_sphere_coords(angular_pixel_coords_f1, self._pixels_per_degree)

        # Frame 2 #
        # --------#

        # sphere to sphere pixel projection

        sphere_coords_f2 = ivy_vision.sphere_to_sphere_coords(
            sphere_coords_f1, agent_rel_mat, [batch_size], self._sphere_img_dims)
        image_var_f2 = image_var_f1

        # to angular pixel coords

        # B x OH x OW x 3
        angular_pixel_coords_f2 = \
            ivy_vision.sphere_to_angular_pixel_coords(sphere_coords_f2, self._pixels_per_degree)

        # constant feature projection

        # B x OH x OW x (3+F)
        projected_coords_f2 = ivy.concatenate([angular_pixel_coords_f2] + [sphere_feat_f1], -1)

        # reshaping to fit quantization dimension requirements

        # B x (OHxOW) x (3+F)
        projected_coords_f2_flat = ivy.reshape(projected_coords_f2,
                                               [batch_size] + [self._sphere_img_dims[0] * self._sphere_img_dims[1]]
                                               + [3 + self._feat_dim])

        # B x (OHxOW) x (3+F)
        image_var_f2_flat = ivy.reshape(image_var_f2,
                                        [batch_size] + [self._sphere_img_dims[0] * self._sphere_img_dims[1]]
                                        + [3 + self._feat_dim])

        # quantize the projection

        # B x N x OH x OW x (3+F)   # B x N x OH x OW x (3+F)
        return ivy_vision.quantize_to_image(
            pixel_coords=projected_coords_f2_flat[..., 0:2],
            final_image_dims=self._sphere_img_dims,
            feat=projected_coords_f2_flat[..., 2:],
            feat_prior=holes_prior,
            with_db=self._with_depth_buffer,
            pixel_coords_var=image_var_f2_flat[..., 0:2],
            feat_var=image_var_f2_flat[..., 2:],
            pixel_coords_prior_var=holes_prior_var[..., 0:2],
            feat_prior_var=holes_prior_var[..., 2:],
            var_threshold=self._var_threshold[:, 0],
            uniform_pixel_coords=uniform_sphere_pixel_coords,
            batch_shape=(batch_size,),
            dev_str=self._dev_str)[0:2]

    # Measurement #
    # ------------#

    def _convert_images_to_omni_observations(self, measurements, uniform_sphere_pixel_coords,
                                             holes_prior, batch_size, num_timesteps, num_cams, image_dims):
        """
        Convert image to omni-directional measurements

        :param measurements: perspective captured images and relative poses container
        :param uniform_sphere_pixel_coords: Uniform  sphere pixel coords *[batch_size, num_timesteps, oh, ow, 3]*
        :param holes_prior: Prior for quantization holes *[batch_size, num_timesteps, oh, ow, 1+f]*
        :param batch_size: Size of batch
        :param num_timesteps: Number of frames
        :param num_cams: Number of cameras
        :param image_dims: Image dimensions
        :return: *[batch_size, n, oh, ow, 3+f]*    *[batch_size, n, oh, ow, 3+f]*
        """

        # coords from all scene cameras wrt world

        images_list = list()
        images_var_list = list()
        cam_rel_poses_list = list()
        cam_rel_poses_cov_list = list()
        cam_rel_mats_list = list()
        validity_mask_list = list()
        for key, item in measurements.to_iterator(leaf_keys_only=True):
            if key == 'img_mean':
                # B x N x 1 x H x W x (3+f)
                images_list.append(ivy.expand_dims(item, 2))
            elif key == 'img_var':
                # B x N x 1 x H x W x (3+f)
                images_var_list.append(ivy.expand_dims(item, 2))
            elif key == 'pose_mean':
                # B x N x 1 x 6
                cam_rel_poses_list.append(ivy.expand_dims(item, 2))
            elif key == 'pose_cov':
                # B x N x 1 x 6 x 6
                cam_rel_poses_cov_list.append(ivy.expand_dims(item, 2))
            elif key == 'cam_rel_mat':
                # B x N x 1 x 3 x 4
                cam_rel_mats_list.append(ivy.expand_dims(item, 2))
            elif key == 'validity_mask':
                validity_mask_list.append(ivy.expand_dims(item, 2))
            else:
                raise Exception('Invalid image key: {}'.format(key))

        # B x N x C x H x W x (3+f)
        images = ivy.concatenate(images_list, 2)

        # B x N x C x H x W x (3+f)
        var_to_project = ivy.concatenate(images_var_list, 2)

        # B x N x C x 6
        cam_to_cam_poses = ivy.concatenate(cam_rel_poses_list, 2)

        # B x N x C x 3 x 4
        cam_to_cam_mats = ivy.concatenate(cam_rel_mats_list, 2)

        # B x N x C x 6 x 6
        cam_to_cam_pose_covs = ivy.concatenate(cam_rel_poses_cov_list, 2)

        # B x N x C x 1
        validity_masks = ivy.concatenate(validity_mask_list, 2) > 0

        # B x N x OH x OW x (3+f)
        holes_prior_var = ivy.ones([batch_size, num_timesteps] + self._sphere_img_dims + [3 + self._feat_dim],
                                   dev_str=self._dev_str) * 1e12

        # reset invalid regions to prior

        # B x N x C x H x W x (3+f)
        images = ivy.where(validity_masks, images,
                           ivy.concatenate((images[..., 0:2], ivy.zeros_like(images[..., 2:], dev_str=self._dev_str)), -1))

        # B x N x C x H x W x (3+f)
        var_to_project = ivy.where(validity_masks, var_to_project,
                                   ivy.ones_like(var_to_project, dev_str=self._dev_str) * 1e12)

        # B x N x OH x OW x (3+f)    # B x N x OH x OW x (3+f)
        return self._frame_to_omni_frame_projection(
            cam_to_cam_poses, cam_to_cam_mats, uniform_sphere_pixel_coords, images[..., 0:3], images[..., 3:],
            cam_to_cam_pose_covs, var_to_project, holes_prior, holes_prior_var, batch_size, num_timesteps, num_cams,
            image_dims)

    # Kalman Filter #
    # --------------#

    def _kalman_filter_on_measurement_sequence(self, prev_fused_val, prev_fused_variance, hole_prior,
                                               hole_prior_var, meas, meas_vars, uniform_sphere_pixel_coords,
                                               agent_rel_poses, agent_rel_pose_covs, agent_rel_mats,
                                               batch_size, num_timesteps):
        """
        Perform kalman filter on measurement sequence

        :param prev_fused_val: Fused value from previous timestamp *[batch_size, oh, ow, (3+f)]*
        :param prev_fused_variance: Fused variance from previous timestamp *[batch_size, oh, ow, (3+f)]*
        :param hole_prior: Prior for holes in quantization *[batch_size, oh, ow, (1+f)]*
        :param hole_prior_var: Prior variance for holes in quantization *[batch_size, oh, ow, (3+f)]*
        :param meas: Measurements *[batch_size, num_timesteps, oh, ow, (3+f)]*
        :param meas_vars: Measurement variances *[batch_size, num_timesteps, oh, ow, (3+f)]*
        :param uniform_sphere_pixel_coords: Uniform sphere pixel co-ordinates *[batch_size, oh, ow, 3]*
        :param agent_rel_poses: Relative poses of agents to the previous step *[batch_size, num_timesteps, 6]*
        :param agent_rel_pose_covs: Agent relative pose covariances *[batch_size, num_timesteps, 6, 6]*
        :param agent_rel_mats: Relative transformations matrix of agents to the previous step
                                *[batch_size, num_timesteps, 3, 4]*
        :param batch_size: Size of batch
        :param num_timesteps: Number of frames
        :return: list of *[batch_size, oh, ow, (3+f)]*,    list of *[batch_size, oh, ow, (3+f)]*
        """

        fused_list = list()
        fused_variances_list = list()

        for i in range(num_timesteps):
            # project prior from previous frame #
            # ----------------------------------#

            # B x OH x OW x (3+F)
            prev_prior = prev_fused_val
            prev_prior_variance = prev_fused_variance

            # B x 3 x 4
            agent_rel_mat = agent_rel_mats[:, i]

            # B x 6
            agent_rel_pose = agent_rel_poses[:, i]

            # B x 6 x 6
            agent_rel_pose_cov = agent_rel_pose_covs[:, i]

            # B x OH x OW x (3+F)   B x OH x OW x (3+F)
            fused_projected, fused_projected_variance = self._omni_frame_to_omni_frame_projection(
                agent_rel_pose, agent_rel_mat, uniform_sphere_pixel_coords, prev_prior[..., 0:2], prev_prior[..., 2:3],
                prev_prior[..., 3:], agent_rel_pose_cov, prev_prior_variance, hole_prior, hole_prior_var, batch_size)

            # reset prior

            # B x OH x OW x (3+F)
            prior = fused_projected
            prior_var = fused_projected_variance

            # per-pixel fusion with measurements #
            # -----------------------------------#

            # extract slice for frame

            # B x OH x OW x (3+F)
            measurement = meas[:, i]
            measurement_variance = meas_vars[:, i]

            # fuse prior and measurement

            # B x 2 x OH x OW x (3+F)
            prior_and_meas = ivy.concatenate((ivy.expand_dims(prior, 1),
                                              ivy.expand_dims(measurement, 1)), 1)
            prior_and_meas_variance = ivy.concatenate((ivy.expand_dims(prior_var, 1),
                                                       ivy.expand_dims(measurement_variance, 1)), 1)

            # B x OH x OW x (3+F)
            low_var_mask = ivy.reduce_sum(ivy.cast(
                prior_and_meas_variance < ivy.expand_dims(hole_prior_var, 1) * self._threshold_var_factor,
                'int32'), 1) > 0

            # B x 1 x OH x OW x (3+F)    B x 1 x OH x OW x (3+F)
            # ToDo: handle this properly once re-implemented with a single scatter operation only
            #  currently depth values are fused even if these are clearly far apart
            fused_val_unsmoothed, fused_variance_unsmoothed = \
                self._fuse_measurements_with_uncertainty(prior_and_meas, prior_and_meas_variance, 1)

            # B x OH x OW x (3+F)
            # This prevents accumulating certainty from duplicate re-projections from prior measurements
            fused_variance_unsmoothed = ivy.where(low_var_mask, fused_variance_unsmoothed[:, 0], hole_prior_var)

            # B x OH x OW x (3+F)
            fused_val = fused_val_unsmoothed[:, 0]
            fused_variance = fused_variance_unsmoothed
            low_var_mask = fused_variance < hole_prior_var

            # B x OH x OW x (3+F)    B x OH x OW x (3+F)
            fused_val, fused_variance = self.smooth(
                fused_val, fused_variance, low_var_mask, self._smooth_mean, self._smooth_kernel_size,
                True, True, batch_size)

            # append to list for returning

            # B x OH x OW x (3+F)
            fused_list.append(fused_val)

            # B x OH x OW x (3+F)
            fused_variances_list.append(fused_variance)

            # update for next time step
            prev_fused_val = fused_val
            prev_fused_variance = fused_variance

        # list of *[batch_size, oh, ow, (3+f)]*,    list of *[batch_size, oh, ow, (3+f)]*
        return fused_list, fused_variances_list

    # Public Functions #
    # -----------------#

    def empty_memory(self, batch_size, timesteps):
        uniform_pixel_coords = \
            ivy_vision.create_uniform_pixel_coords_image(self._sphere_img_dims, [batch_size, timesteps],
                                                         dev_str=self._dev_str)[..., 0:2]
        empty_memory = {
            'mean': ivy.concatenate(
                [uniform_pixel_coords] +
                [ivy.ones([batch_size, timesteps] + self._sphere_img_dims + [1], dev_str=self._dev_str)
                 * self._sphere_depth_prior_val] +
                [ivy.ones([batch_size, timesteps] + self._sphere_img_dims + [self._feat_dim], dev_str=self._dev_str)
                 * self._sphere_feat_prior_val], -1),
            'var': ivy.concatenate(
                [ivy.ones([batch_size, timesteps] + self._sphere_img_dims + [2], dev_str=self._dev_str)
                 * self._ang_pix_prior_var_val] +
                [ivy.ones([batch_size, timesteps] + self._sphere_img_dims + [1], dev_str=self._dev_str)
                 * self._depth_prior_var_val] +
                [ivy.ones([batch_size, timesteps] + self._sphere_img_dims + [self._feat_dim], dev_str=self._dev_str)
                 * self._feat_prior_var_val], -1)
        }
        return ESMMemory(**empty_memory)

    def smooth(self, fused_val, fused_variance, low_var_mask, smooth_mean,
               smooth_kernel_size, variance_mode, fix_low_var_pixels, batch_size):
        variance_mode = True
        var_for_smoothing = fused_variance

        # image smoothing
        if smooth_mean:

            # pad borders
            pad_size = int(smooth_kernel_size / 2)
            fused_val_pad = ivy_vision.pad_omni_image(fused_val, pad_size, self._sphere_img_dims)
            fused_variance_pad = ivy_vision.pad_omni_image(var_for_smoothing, pad_size, self._sphere_img_dims)
            expanded_sphere_img_dims = [item + 2 * pad_size for item in self._sphere_img_dims]

            # reshape for dilation and erosion
            fused_val_pad_flat = ivy.reshape(fused_val_pad[..., 2:], [batch_size] + expanded_sphere_img_dims +
                                             [1 + self._feat_dim])
            fused_var_pad_flat = ivy.reshape(fused_variance_pad[..., 2:], [batch_size] + expanded_sphere_img_dims +
                                             [1 + self._feat_dim])

            if variance_mode:
                smoothed_fused_val_flat, smoothed_fused_var_flat = \
                    ivy_vision.smooth_image_fom_var_image(fused_val_pad_flat, fused_var_pad_flat, smooth_kernel_size,
                                                          ivy.array([1.] * (1 + self._feat_dim), dev_str=self._dev_str))
            else:
                smoothed_fused_val_flat, smoothed_fused_var_flat = \
                    ivy_vision.weighted_image_smooth(fused_val_pad_flat, 1 - fused_var_pad_flat, smooth_kernel_size)

            # reshape to image dims
            smoothed_fused_val = ivy.reshape(smoothed_fused_val_flat, [batch_size] + self._sphere_img_dims +
                                             [1 + self._feat_dim])
            smoothed_fused_val = ivy.concatenate((fused_val[..., 0:2], smoothed_fused_val), -1)

            # replace temporary zeros with their prior values
            # This ensures that the smoothing operation only changes the values for regions of high variance
            if fix_low_var_pixels:
                fused_val = ivy.where(low_var_mask, fused_val, smoothed_fused_val)
            else:
                fused_val = smoothed_fused_val

        return fused_val, fused_variance

    # Main call method #
    # -----------------#

    def _forward(self, obs: ESMObservation, memory: ESMMemory = None, batch_size=None, num_timesteps=None,
                 num_cams=None, image_dims=None):
        """
        Perform ESM update step.

        :param obs: Observations
        :type obs: ESMObservation
        :param memory: Memory from the previous time-step, uses internal parameter if None.
        :type memory: ESMMemory, optional.
        :param batch_size: Size of batch, inferred from inputs if None.
        :type batch_size: int, optional
        :param num_timesteps: Number of timesteps, inferred from inputs if None.
        :type num_timesteps: int, optional
        :param num_cams: Number of cameras, inferred from inputs if None.
        :type num_cams: int, optional
        :param image_dims: Image dimensions of captured images, inferred from inputs if None.
        :type image_dims: sequence of ints, optional
        :return: New memory of type ESMMemory
        """

        # get shapes
        img_meas = (next(iter(obs.img_meas.values()))).img_mean
        if batch_size is None:
            batch_size = int(img_meas.shape[0])
        if num_timesteps is None:
            num_timesteps = int(img_meas.shape[1])
        if num_cams is None:
            num_cams = len(obs.img_meas.values())
        if image_dims is None:
            image_dims = list(img_meas.shape[2:4])

        # get only previous memory

        # extract from memory #
        # --------------------#

        if memory:
            prev_mem = memory[:, -1]
        elif self._stateful and self._memory is not None:
            prev_mem = self._memory[:, -1]
        else:
            prev_mem = self.empty_memory(batch_size, 1)[:, -1]

        # holes prior #
        # ------------#

        # B x N x OH x OW x 1
        self._sphere_depth_prior = \
            ivy.ones([batch_size, num_timesteps] + self._sphere_img_dims + [1], dev_str=self._dev_str) * \
            self._sphere_depth_prior_val

        # B x N x OH x OW x F
        self._sphere_feat_prior = \
            ivy.ones([batch_size, num_timesteps] + self._sphere_img_dims + [self._feat_dim], dev_str=self._dev_str) * \
            self._sphere_feat_prior_val

        # B x N x OH x OW x (1+F)
        holes_prior = ivy.concatenate([self._sphere_depth_prior] + [self._sphere_feat_prior], -1)

        # holes prior variances #
        # ----------------------#

        # B x N x OH x OW x 2
        sphere_ang_pix_prior_var = \
            ivy.ones([batch_size, num_timesteps] + self._sphere_img_dims + [2], dev_str=self._dev_str) * self._ang_pix_prior_var_val

        # B x N x OH x OW x 1
        sphere_depth_prior_var = \
            ivy.ones([batch_size, num_timesteps] + self._sphere_img_dims + [1], dev_str=self._dev_str) * self._depth_prior_var_val

        # B x N x OH x OW x F
        sphere_feat_prior_var = \
            ivy.ones([batch_size, num_timesteps] + self._sphere_img_dims + [self._feat_dim], dev_str=self._dev_str) * \
            self._feat_prior_var_val

        # B x N x OH x OW x (3+F)
        holes_prior_var = ivy.concatenate([sphere_ang_pix_prior_var] +
                                          [sphere_depth_prior_var] +
                                          [sphere_feat_prior_var], -1)

        # variance threshold #
        # -------------------#

        # B x N x (3+F) x 1
        var_threshold_min = ivy.tile(
            ivy.reshape(ivy.stack([self._min_ang_pix_var]*2 +
                                  [self._min_depth_var] +
                                  [self._min_feat_var] * self._feat_dim),
                        [1, 1, 3 + self._feat_dim, 1]), [batch_size, num_timesteps, 1, 1])
        var_threshold_max = ivy.tile(ivy.reshape(ivy.stack([self._ang_pix_var_threshold] * 2 +
                                                           [self._depth_var_threshold] +
                                                           [self._feat_var_threshold] * self._feat_dim),
                                                 [1, 1, 3 + self._feat_dim, 1]), [batch_size, num_timesteps, 1, 1])
        self._var_threshold = ivy.concatenate((var_threshold_min, var_threshold_max), -1)

        # measurements #
        # -------------#

        # B x N x OH x OW x 3
        uniform_sphere_pixel_coords = ivy_vision.create_uniform_pixel_coords_image(
            self._sphere_img_dims, (batch_size, num_timesteps), dev_str=self._dev_str)

        # B x N x OH x OW x (3+F),    B x N x OH x OW x (3+F)
        meas_means, meas_vars = self._convert_images_to_omni_observations(
            obs.img_meas, uniform_sphere_pixel_coords, holes_prior, batch_size, num_timesteps, num_cams, image_dims)

        # filtering #
        # ----------#

        # list of B x OH x OW x (3+F),    list of B x OH x OW x (3+F)
        fused_measurements_list, fused_variances_list = \
            self._kalman_filter_on_measurement_sequence(
                prev_mem.mean, prev_mem.var, holes_prior[:, 0], holes_prior_var[:, 0], meas_means, meas_vars,
                uniform_sphere_pixel_coords[:, 0], obs.control_mean, obs.control_cov, obs.agent_rel_mat,
                batch_size, num_timesteps)

        # new variance #
        # -------------#

        # B x N x OH x OW x (3+F)
        fused_variance = ivy.concatenate([ivy.expand_dims(item, 1) for item in fused_variances_list], 1)

        # variance clipping

        # B x N x OH x OW x 1
        fused_depth_variance = ivy.clip(fused_variance[..., 0:1], self._min_depth_var, self._depth_prior_var_val)

        # B x N x OH x OW x 3
        fused_feat_variance = ivy.clip(fused_variance[..., 1:], self._min_feat_var, self._feat_prior_var_val)

        # B x N x OH x OW x (3+F)
        fused_variance = ivy.concatenate([fused_depth_variance] + [fused_feat_variance], -1)

        # new mean #
        # ---------#

        # B x N x OH x OW x (3+F)
        fused_measurement = ivy.concatenate([ivy.expand_dims(item, 1) for item in fused_measurements_list], 1)

        # value clipping

        # B x N x OH x OW x 2
        fused_pixel_coords = fused_measurement[..., 0:2]

        # B x N x OH x OW x 1
        fused_depth = ivy.clip(fused_measurement[..., 2:3], self._min_depth, self._max_depth)

        # B x N x OH x OW x 3
        fused_feat = fused_measurement[..., 3:]

        # B x N x OH x OW x (3+F)
        fused_measurement = ivy.concatenate([fused_pixel_coords] + [fused_depth] + [fused_feat], -1)

        # update memory #
        # --------------#

        # B x N x OH x OW x (3+F),    B x N x OH x OW x (3+F)
        self._memory = ESMMemory(mean=fused_measurement, var=fused_variance)

        # return #
        # -------#

        return self._memory



# lint as: python3
# Copyright 2021 The Ivy Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License..
# ==============================================================================
from distutils.core import setup
import setuptools

setup(name='ivy-memory',
      version='1.1.6',
      author='Ivy Team',
      author_email='ivydl.team@gmail.com',
      description='End-to-end memory modules for deep learning developers, written in Ivy.',
      long_description="""# What is Ivy Memory?\n\nIvy memory provides differentiable memory modules, including learnt
      modules such as Neural Turing Machines (NTM), but also parameter-free modules such as End-to-End Egospheric
      Spatial Memory (ESM). Ivy currently supports Jax, TensorFlow, PyTorch, MXNet and Numpy.
      Check out the [docs](https://ivy-dl.org/memory) for more info!""",
      long_description_content_type='text/markdown',
      url='https://ivy-dl.org/memory',
      project_urls={
            'Docs': 'https://ivy-dl.org/memory/',
            'Source': 'https://github.com/ivy-dl/memory',
      },
      packages=setuptools.find_packages(),
      install_requires=['ivy-core', 'ivy-mech', 'ivy-vision'],
      classifiers=['License :: OSI Approved :: Apache Software License'],
      license='Apache 2.0'
      )
